;Project homepage: www.HomoFaciens.de
;Program indicates number of revolutions of an electric motor, using optical sensors. For details visit the hompage of the project.
;Program code written for ATmega8A microcontroller
;Copyright (C) 2011 Norbert Heinz
;
;This program is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation version 3 of the License.
;This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
;You should have received a copy of the GNU General Public License along with this program if not, see http://www.gnu.org/licenses/.

;Version 0.1

.include "m8def.inc"

.def temp01  = r16
.def temp02  = r17
.def TimeWait = r18         ;count number of interrupts
.def OverflowCount = r19    ;Number of overflows of PulseCount
.def ControlBits = r20      ;bit 0 - previous state of photo sensor
                            ;bit 1 - state of hardware jumper
.def PulseCount = r21       ;number of pulses
.def OverflowLimit01 = r22  ;different overflow limits for method 1 and method 2
.def OverflowLimit02 = r23
.def PulseAdd = r24         ;different multiplicators for PulseCount used at method 1 and method 2



.org 0x0000
        rjmp    main                    ;Reset Handler

.org OC1Aaddr
        rjmp    timer1_compare          ;Timer Compare Handler

main:
        ldi     temp01, LOW(RAMEND)     ;initialize stackpointer
        out     SPL, temp01
        ldi     temp01, HIGH(RAMEND)
        out     SPH, temp01

        ;initialize ports
        ldi     temp01, 0b11111111      ;Set all pins of port B as output
        out     DDRB, temp01

        ldi     temp01, 0b11111111      ;Set all pins of Port D as output
        out     DDRD, temp01

        ldi     temp01, 0b00000000      ;Set all pins of Port C as input
        out     DDRC, temp01
        ldi     temp01, 0b00000010      ;Pullup resistance at pin 1
        out     PORTC, temp01

        ;initialize timer
        ldi     temp01, high( 40000 - 1 )     ;Compare value = 40,000 pulses = 0.01 seconds until interrupt
        out     OCR1AH, temp01
        ldi     temp01, low( 40000 - 1 )
        out     OCR1AL, temp01
        ldi     temp01, ( 1 << WGM12 ) | ( 1 << CS10 ) ;CTC mode on, divisor 1
        out     TCCR1B, temp01

        ldi     temp01, 1 << OCIE1A                    ;OCIE1A: interrupt at timer compare
        out     TIMSK, temp01

        clr     TimeWait
        clr     OverflowCount
        clr     ControlBits

        ;get initial state of disc
        sbis    PINC, 0
        ori     ControlBits, 0b00000001


        sei     ;enable global interrupt

loop:

        ;check hardware jumper
        sbic    PINC, 1
        ori     ControlBits, 0b00000010     ;jumper open, show binary value
        sbis    PINC, 1
        andi    ControlBits, 0b11111101     ;jumper closed, show LED bar

        ldi     OverflowLimit01, 100        ;count 100 pulses until overflow (method 1, LED bar)
        ldi     OverflowLimit02, 3          ;two red LEDs show the number of overflows (0-3)
        ldi     PulseAdd, 1
        sbrs    ControlBits, 1
        ldi     OverflowLimit01, 255        ;count 255 pulses until overflow (method 2, binary value)
        sbrs    ControlBits, 1
        ldi     OverflowLimit02, 15         ;four high LEDs encode numbers 0-15 (method 2, binary value)
        sbrs    ControlBits, 1
        ldi     PulseAdd, 5                 ;PulseCount * 5 gives the number of revolutions per minute (method 2, binary value)
        ;revolution indication 2


        ;check photo sensor to count pulses
        sbis    PINC, 0
        rjmp    skip01
        sbrc    ControlBits, 0
        rjmp    PulseCountEnd
        ori     ControlBits, 0b00000001        ;photo sensor status changed from 0 to 1
        cp      PulseCount, OverflowLimit01
        breq    PulseOverflow
        add     PulseCount, PulseAdd
        rjmp    PulseCountEnd

skip01:
        sbrs    ControlBits, 0
        rjmp    PulseCountEnd
        andi    ControlBits, 0b11111110      ;photo sensor status changed from 1 to 0
        cp      PulseCount, OverflowLimit01
        breq    PulseOverflow
        add     PulseCount, PulseAdd
        rjmp    PulseCountEnd

PulseOverflow:
        cpi     OverflowCount, OverflowLimit02
        brsh    PulseCountEnd
        clr     PulseCount
        inc     OverflowCount

PulseCountEnd:

        rjmp    loop



;++++++++++++++++++++++++++++++++++ Timer interrupt +++++++++++++++++++++++++++++++++++++++

timer1_compare:                       ;Timer 0 Overflow Handler
        push    temp01                ;save temp01
        in      temp01, SREG          ;save status register
        push    temp01

        ;timer overflow while running at 4MHz:
        ;4,000,000 / 40,000 = 100 interrupts per second
        ;one overflow each 0.01 second
        ;after 50 timer interrupts (=0.5seconds) the number of pulses is calulated
        ;disc triggers 24 pulses per revolution
        ;PulseCount / 24 *  120 = PulseCount * 5 gives the number of revolutions per minute
        ;Method 1: each LED equals 50 rpm, red LEDs indicate number of overflows as binary value (0-2000rpm)
        ;Method 2: rpm is given as binary value (0-4096rpm)
        inc     TimeWait
        cpi     TimeWait, 50
        brlo    SkipOutput

        sbrs    ControlBits, 1
        rjmp    OutputMethod02

        ;display method 1, statur bar with 50rpm per LED.
        ;the two red LEDs display the number of overflows as binary value
        ldi     temp01, 0b00000000
        ldi     temp02, 0b11111100
        cpi     PulseCount, 90
        brsh    SetOutputMethod01

        ldi     temp01, 0b00000000
        ldi     temp02, 0b11111110
        cpi     PulseCount, 80
        brsh    SetOutputMethod01

        ldi     temp01, 0b00000000
        ldi     temp02, 0b11111111
        cpi     PulseCount, 70
        brsh    SetOutputMethod01

        ldi     temp01, 0b10000000
        cpi     PulseCount, 60
        brsh    SetOutputMethod01

        ldi     temp01, 0b11000000
        cpi     PulseCount, 50
        brsh    SetOutputMethod01

        ldi     temp01, 0b11100000
        cpi     PulseCount, 40
        brsh    SetOutputMethod01

        ldi     temp01, 0b11110000
        cpi     PulseCount, 30
        brsh    SetOutputMethod01

        ldi     temp01, 0b11111000
        cpi     PulseCount, 20
        brsh    SetOutputMethod01

        ldi     temp01, 0b11111100
        cpi     PulseCount, 10
        brsh    SetOutputMethod01

        ldi     temp01, 0b11111110
        cpi     PulseCount, 1
        brsh    SetOutputMethod01

        ldi     temp01, 0b11111111

SetOutputMethod01:

        ;the two red LEDs show the number of overflows as binary value
        cpi     OverflowCount, 0
        breq    EndOverflowOutput
        andi    temp02, 0b11111011

        cpi     OverflowCount, 1
        breq    EndOverflowOutput
        ori     temp02, 0b00000100
        andi    temp02, 0b11110111

        cpi     OverflowCount, 2
        breq    EndOverflowOutput
        andi    temp02, 0b11111011

EndOverflowOutput:

        rjmp    SetOutput

OutputMethod02:
        ;display method 2: number of revolutions per minute shown as binary value
        ldi     temp01, 0b11111111
        eor     PulseCount, temp01     ;toggle bits, because 1 sets LED is OFF, 0 sets LED ON
        eor     OverflowCount, temp01  ;toggle bits
        mov     temp01, PulseCount
        mov     temp02, OverflowCount

SetOutput:
        out     PORTB, temp01
        out     PORTD, temp02

        clr     PulseCount
        clr     OverflowCount
        clr     TimeWait

SkipOutput:

        pop     temp01
        out     SREG, temp01          ;restore status register
        pop     temp01                ;restore temp01

        reti